package PIPC;

# Protocol codes (codes 128 through 159 are unused in the ASCII character set)
my $PIPC_PTLID = 128;   # Protocol identifier
my $PIPC_ERROR = 129;   # Error message
my $PIPC_CMPID = 130;   # PID of the component application
my $PIPC_ENDOS = 131;   # End of session .Signals the client that the connection should be terminated.
my $PIPC_CUSTM = 140;   # Custom protocol codes start from 140. See the GetPtlCustomCodeRange() function.

my $PIPC_NEWLN = "%%";  # LF's in the tsrings sent to the component should be replaced with this sequence.

# Global vars
my $g_bClientReady = 0; # Clent is ready to receive component's data
my $g_bTrace = 0;       # Signal handlers can only be global




sub new {
    my $class = shift;
    my $self = {};

    ${ $self->{ hHooks } }{ TERM } = \&TermHandler;
    ${ $self->{ hHooks } }{ TSTP } = \&PauseHandler;
    ${ $self->{ hHooks } }{ CONT } = \&ContHandler;
    ${ $self->{ hHooks } }{ USR1 } = \&Usr1Handler;

    bless $self;
    return $self;
}


# A PIPC-derived class will need to know $PIPC_CUSTM if it wants to
# use custom protocol codes. Since PIPC global vars are invisible
# outside the PIPC.pm module, and I don't wanna add another hash 
# table to the PIPC package at this point, the only reasonable way
# to obtain $PIPC_CUSTM is thru a static method below.
sub GetPtlCustomCodeRange {
    my $class = shift;
    my $self = {};
    bless $self;
    return $PIPC_CUSTM;
}


sub Connect {
    my $self = shift;

    $self->HookSignal( 'TERM' );
    $self->HookSignal( 'TSTP' );
    $self->HookSignal( 'CONT' );
    $self->HookSignal( 'USR1' );

    # Pass component PID to the client
    $self->SendMessage( $PIPC_CMPID, $$ );

    # Wait for the client
    while( !$g_bClientReady ) {
        select( undef, undef, undef, 0.1 );
    }

    $self->UnhookSignal( 'USR1' );
}


sub Disconnect {
    my $self = shift;
    $self->SendMessage( $PIPC_ENDOS );
}


sub SendMessage {
    my $self = shift;
    my $string = sprintf( "%c%c", $PIPC_PTLID, shift );
    $string .= "$_[0]$_[1]";  # can't use @_, since it inserts a whitespace between the two args

    # Replace newline chars with $PIPC_NEWLN, so the component will be able to get the message
    # in one piece. Having done that, it will translate PIPC_NEWLNs back to newline chars.
    $string =~ s/\n/$PIPC_NEWLN/g;

    print "$string\n";
}


sub PrintError {
    my $self = shift;
    $self->SendMessage( $PIPC_ERROR, shift );
}


sub EnableTracing {
    my $self = shift;
    $g_bTrace = ( @_ ) ? shift : 0;
}


sub HookSignal {
    my $self = shift;
    my $sig = shift;

    ${ $self->{ hOldHooks } }{ $sig } = $SIG{ $sig };
    $SIG{ $sig } = ${ $self->{ hHooks } }{ $sig };
}


sub UnhookSignal {
    my $self = shift;
    my $sig = shift;
    
    $SIG{ $sig } = ${ $self->{ hOldHooks } }{ $sig };
}


sub TRACE {
    print STDERR "\e[1mTRACE[ prepro.pl ]:\e[0m $_[0]\r\n" if( $g_bTrace );
}


# Signal handlers

sub PauseHandler {
    TRACE( "Signal intercept: SIGTSPS" );
    # Perl automatically unhooks the handler of an intercepted signal
    $SIG{ TSTP } = \&PauseHandler;  
    sleep;
}


sub ContHandler {
    TRACE( "Signal intercept: SIGCONT" );
    $SIG{ CONT } = \&ContHandler;
}

sub TermHandler {
    TRACE( "Signal intercept: SIGTERM" );
    TRACE( "Terminating." );
    exit 0;
}

sub Usr1Handler {
    TRACE( "Signal intercept: SIGUSR1" );
    $g_bClientReady = 1;
}


1;  # to satisfy the compiler

