#!/bin/env mwperl
do "$ENV{MWHOME}/lib/la/Object_Module.pl";
do "$ENV{MWHOME}/lib/la/Symbol_Table.pl";
# Holds a symbol table, a name and whether it's used.
# This class is for internal use.
use strict;
package Named_Sym_Table; {

sub new {
    my $self = bless {}, shift;
    $self->{name} = shift;
    $self->{table} = shift;
    $self->{dynamic} = shift;
    # Is this object file used?
    $self->{used} = 0;
    return $self;
}

} # End Named_Sym_Table


# Represents an archive file.
package Archive_Module; @Archive_Module::ISA = qw(Object_Module B_Status_Aware); {

sub new {
    # Note that we don't call Object_Module's constructor
    # this is because we want to inherit the interface but
    # not the structure (in C++ we would do this with a common
    # abstruct base class).
    my ($type, $data, $bstatus, $lib_obj) = (shift, shift, shift, shift);
    my $self = Link_Line_Atom::new($type, $data);

    $self->set_b_status($bstatus);

    # If this is true we don't check whether we need to add 
    # this lib to the link line, otherwise we check.
    $self->{_always_added} = 0;

    # If this is true, the archive itself decided that it is needed
    $self->{_needed} = 0;

    $self->{_lib_obj} = $lib_obj;
    $self->_process_a_file();

    return $self;
}

sub resolves_symbol {
    my ($self, $symbol) = @_;
    if( not defined $symbol ) { return ""; }

    foreach my $st (@{$self->{_sym_tables}}) {
	if( $st->{table}->is_defined($symbol) ) {
	    return $self->name().":".$st->{name};
	}
    }
    return "";
}

# Returns: is this symbol supplied by this module?
# The object file that provides the symbol must be used.
sub provides_symbol {
    my ($self, $symbol) = @_;
    return if not defined $symbol;
    foreach my $st (@{$self->{_sym_tables}}) {
	if( $st->{used} and $st->{table}->is_defined($symbol) ) {
	    return 1;
	}
    }
    return ;
}

# Add only the needed object files (to both symbol tables)
sub add_self {
    my ($self, $parser) = @_;

    # Log apperence of library on the link line.
    $parser->saw_library($self->{_lib_obj}->{stripped_name}, $self);

    # all system and empty libs get added and we don't need to analyze them
    if( $self->is_system() || $self->is_empty_stripped() ) { 
	$self->{_needed} = 1;
	return; 
    }
   
    # Clear used bits so we get a fresh look on things
    foreach my $st (@{$self->{_sym_tables}}) {
	$st->{used} = 0;
    }
    $self->{_needed} = 0;

    # Keep looping until a whole round is done without
    # adding an object file. This is done so that if an
    # inserted object requires a symbol that was defined
    # in an earlier object it will be found.
    my $changed = 1;

    # Keep track of all symbols that were already defined
    my @old_syms = ();

    while( $changed ) {
	$changed = 0;
	# iterate over all .o files in the archive
	foreach my $st (@{$self->{_sym_tables}}) {
	    next if $st->{used}; # already inserted

	    # In linux we need to keep track of defined symbols that 
	    # are redefined
	    my $old_s;
	    if( $ENV{MWOS} eq 'linux' ) {
		$old_s = [];
	    }

	    # Note we do a conditional merge
	    my $reason;
	    my $ret;
	    if( $st->{dynamic} ) {		    
		$ret = $parser->merge_dynamic($st->{table}, 1, $old_s, \$reason);
	    }
	    else {
		$ret = $parser->merge_static($st->{table}, 1, $old_s, \$reason);
	    }
	    if ($ret) {
		if ($ret eq 1) {
		    $st->{used} = 1;
		    print STDERR "Adding ",($st->{dynamic} ? "shared library" : "object file")," $st->{name} from archive $self->{_lib_obj}->{stripped_name} because it resolves symbol $reason\n"
			if $linker_analyzer::Full_Trace;
		    $parser->require_libs_from_pragma_comment_lib( $st->{table} );
		    push @old_syms, @$old_s if $old_s;
		    $changed = 1; # file added in this round
		} else {
		    print STDERR "Keeping archive $self->{_lib_obj}->{stripped_name} because object file $st->{name} promotes common symbol $reason\n"
			if $linker_analyzer::Full_Trace;
		}
		if( ! $self->{_needed} ) {
		    $self->B_Status_Aware::add_self($parser);
		    $self->{_needed} = 1;
		}
	    }
	}
    }
    
    # We need to do this if any one of the object files in the
    # archive isn't used so it's probably better to just do it
    # anyway and let unneeded_lib find out it has nothing to do.
    $parser->unneeded_lib($self);

    return @old_syms;
}

sub set_added {
    my ($self, $uflag) = @_;
    
    $self->{_always_added} = 1;
    
    $self->prepend($uflag);
}

# An archive is used int the final link if one of its object
# files is used.
sub is_added {
    return 1 if $LA_Flags::Do_Not_Remove;
    my $self = shift;
    if( $self->{_always_added} ) {
	return 1;
    }
    $self->{_needed} = 1 if $self->is_system() || $self->is_empty_stripped();
    return $self->{_needed};
}

sub flush {
    my $self = shift;
    $self->{_needed} = 0;
    $self->{_always_added} = 0;
}

sub get_unused_defined {
    my $self = shift;
    my @ret = ();
    foreach my $st (@{$self->{_sym_tables}}) {
	if( not $st->{used} ) {
	    push(@ret, $st->{table}->get_defined());
	}
    }
    return @ret;
}

sub is_empty_stripped {
    my $self = shift;
    return ($self->{_lib_obj}->{is_stripped} && $self->{_lib_obj}->{is_empty});
}

sub is_system {
    my $self = shift;
    return $self->{_lib_obj}->{is_system};
}

sub get_text {
    my $self = shift;
    if( $self->is_added() )  {
	my $ret = $self->B_Status_Aware::get_text();
	$ret .= $self->Object_Module::get_text();
	return $ret;
    }
    return "";
}

sub stripped_name {
    my $self = shift;
    return $self->{_lib_obj}->{stripped_name};
}


# Description:
# process a .a file.
sub _process_a_file
{
    my $self = shift;
    my $lib_obj = $self->{_lib_obj};
    my ($ty, $info);

    if (  $self->is_system() || $self->is_empty_stripped() ) {
	$self->{_always_added} = 1;
	return;
    }

    if ( not defined $lib_obj->{"path"} ) { 
	print STDERR "WARNING: ",$self->name()," Got an archive lib object with no defined path!\n";
	return;
    }

    my ($o_file) = "";
    my $is_dynamic;

    if ( $linker_analyzer::Full_Trace ) {
	print STDERR "processing archive : $lib_obj->{path}\n";
	if ( $lib_obj->{"is_default"} ) { 
	    print STDERR "$lib_obj->{path} is a default lib\n";
	}
    }
    
    my $nm_data = $lib_obj->get_nm_data_file_handle();
    my @data = ();
    my @sym_tbls = ();
    while ( <$nm_data> ) {
	my $line = $_;
	chomp($line);

	my ($ty, $info) = split(" ", $line);
	# handle the header
	if (  $ty eq "H" or $ty eq "S")  {
	    if( @data ) {
		my $table = $self->_build_sym_table(@data);
		
		my $named_st = Named_Sym_Table->new($o_file, $table, $is_dynamic);
		
		push(@sym_tbls, $named_st);
	    }
	    $o_file = $info;
	    $is_dynamic = ($ty eq "S");
	    @data = ();
	    next;
	}

	push (@data, $line);
    }
    close($nm_data);

    # re-do last O file
    if( @data ) {
	my $table = $self->_build_sym_table(@data);
	my $named_st = Named_Sym_Table->new($o_file, $table, $is_dynamic);
	push(@sym_tbls, $named_st);
    }

    # Keep the symbol tables in an array because we want 
    # to maintain their order.
    $self->{_sym_tables} = \@sym_tbls;
}   

} # Archive_Module

# An archive that was added using the -l flag
package Archive_Module_L; @Archive_Module_L::ISA = qw(Archive_Module); {

sub new {
    my ($type, $flag, $file) = (shift, shift, shift);
    my $self = Archive_Module::new($type, $file, @_);

    $self->{_flag} = $flag; # save how it looks on the command line
    
    return $self;
}

sub get_text {
    my $self = shift;
    if( $self->is_added() ) {
	my $ret = $self->B_Status_Aware::get_text();

	# if the library was found via the .lib file add the directory
	# to the lib path.
	if( $self->{_lib_obj} and $self->{_lib_obj}->{add_dash_L} ) {
	    my $dir = $self->{_lib_obj}->{add_dash_L};
	    if( not $Link_Parser::added_lib_dirs{$dir} ) {
		$ret .= " -L".$self->{_lib_obj}->{add_dash_L}." ";
		$Link_Parser::added_lib_dirs{$dir} = 1;
	    }   
	}

	$ret .= "-l".$self->{_flag};
	return $ret;
    }
    return "";
}

}
1;
