#!/bin/env mwperl
# Represents a directed graph
# Please don't use stuff that begins with an underscore
# it's private.

# A directed graph's node.
use strict;
package DGraph_Node;
use Carp;
sub new {
    my $self = bless {}, shift;
    $self->{_object} = shift;
    $self->{_tag} =  0; # has this node been visited?
    $self->{_in} =  {}; # incoming edges
    $self->{_out} = {}; # outgoing edges
    return $self;
}

# Has the element been tagged.
sub tagged {
    my $self = shift;
    return $self->{_tag}? 1 : 0;
}

# Set tag, default is true.
sub tag {
    my $self = shift;
    $self->{_tag} = @_? shift : 1;
}

# How many un-tagged incoming edges are there?
sub level {
    my $self = shift;
    return scalar $self->incoming_untagged();
}

sub incoming_untagged {
    my $self = shift;
    return grep { not $_->tagged() } values %{$self->{_in}};
}

sub outgoing_nodes {
    my $self = shift;
    return values %{$self->{_out}};
}

sub add_outgoing {
    my $self = shift;
    for my $other (@_) {
	if( not isa_DG_node($other) ) {
	    carp "Trying to add an edge to a bad object ($other)";
	    next;
	}
	$self->{_out}->{$other} = $other;
	$other->{_in}->{$self} = $self;
    }
}

sub remove_outgoing {
    my $self = shift;
    for my $other (@_) {
	next unless $self->{_out}->{$other};
	delete $self->{_out}->{$other};
	delete $other->{_in}->{$self};
    }
}

sub get_object {
    my $self = shift;
    return $self->{_object};
}

sub isa_DG_node {
    eval { shift()->isa("DGraph_Node"); };
}


# The Graph itself
package Directed_Graph;
sub new {
    my $self = bless {}, shift;
    $self->{_graph} = {};
    return $self;
}

# Gets a list of objects and returns a list of corosponding nodes 
# in the same order, the nodes are created if they don't yet exist.
sub get_nodes {
    my $self = shift;
    my @ret;
    for my $obj(@_) {
	my $node = $self->{_graph}->{$obj};
	unless( $node ) {
	    $node = DGraph_Node->new($obj);
	    $self->{_graph}->{$obj} = $node;
	}
	push(@ret, $node);
    }
    return @ret;
}
  
sub add_edge {
    my $self = shift;
    die "Need two nodes to make an edge" if @_ != 2;

    my ($from, $to) = $self->get_nodes(@_);

    $from->add_outgoing($to);
}

# Get the untagged node with the lowest number of untagged incoming nodes.
sub get_minimal {
    my $self = shift;
    my @untagged = grep { not $_->tagged() } values %{$self->{_graph}};
    my @srt = sort {$a->level() <=> $b->level()} @untagged;
    return $srt[0];
}

# "Untag" all the nodes in the graph.
sub clear_all {
    my $self = shift;
    $_->tag(0) for values %{$self->{_graph}};
}

# Turn the graph into a dag, remove all cycles.
sub dagify {
    my $self = shift;
    while( my $head = $self->get_minimal() ) {
	foreach my $in ( $head->incoming_untagged() ) {
	    print STDERR "Located a cycle in the shared object dependency tree,",
	    "removing the dependency btween ",$in->get_object()->name(), " and ",
	    $head->get_object()->name(),"!\n" if $linker_analyzer::Partial_Trace;

	    $in->remove_outgoing($head);
	}
	$head->tag(1);
    }
    $self->clear_all();
}

sub get_dependents {
    my ($self, $obj) = @_;
    my $node = $self->{_graph}->{$obj};
    return unless $node;
    # Get untagged incoming nodes
    my @preds = $node->incoming_untagged();
    # return objects
    return map {$_->get_object()} @preds;
}

sub tag_node { 
    my ($self, $obj) = @_;
    my $node = $self->{_graph}->{$obj};
    $node->tag(1) if $node;
}



1;
