#!/bin/env mwperl

# An element in a linked list.
# Derive from this class to put a class in a linked list.
# Stuff begining with an underscore is private.
use strict;
package Linked_List_Elem; {

# Insert a LLE in front of this one
# Return: if the appended ellement is valid it is returned otherwise $self
# is returned, this allows us to write $x = $x->append($y) and know we still
# have a valid LLE
sub append {
    my ($self, $other) = @_;

    return $self unless _isaLLE($other); 

    my $next = $self->{_next};

    $self->{_next} = $other;
    $other->{_prev} = $self;

    $other->{_next} = $next;
    if( $next ) {
	$next->{_prev} = $other;
    }
    return $other;
}

# Insert a LLE before this one
# Return: See append()
sub prepend {
    my ($self, $other) = @_;

    return $self unless _isaLLE($other);

    my $prev = $self->{_prev};

    $other->{_next} = $self;
    $self->{_prev} = $other;
    
    $other->{_prev} = $prev;
    if( $prev ) {
	$prev->{_next} = $other;
    }
    return $other;
}

# Remove this LLE from the list
# returns the element for ease of use.
sub remove {
    my $self = shift;
    my $next = $self->{_next};
    my $prev = $self->{_prev};
    
    if( $next ) {
	$next->{_prev} = $prev;
    }
    if( $prev ) {
	$prev->{_next} = $next;
    }
    
    # clear refs, just to be on the safe side.
    $self->{_next} = $self->{_prev} = "";

    return $self;
}

sub next_elem {
    my $self = shift;
    return $self->{_next};
}

sub previous_elem {
    my $self = shift;
    return $self->{_prev};
}

# This checks if something is a Linked_List_Elem we use
# it to make sure we don't get something which isn't a blessed
# LLE and attempt to insert it into the list.
# This should be rather robust and deal even with the undef value.
sub _isaLLE {
    my $x = shift;
    my $ret = eval { $x->isa("Linked_List_Elem"); };
    if( not $ret and $linker_analyzer::Partial_Trace) {
	print STDERR "ERROR! Tried to add a bad object to a linked list! ($x)\n";
    }
    return $ret;
}

} # Linked_List_Elem

# Represents an item on the link line.
# Every atom can do a minimum of things.
# 1) Give a textual representation of itself which can
#    be added to a native unix linker's command line.
# 2) Get added to the link line data object (in the
#     trivial case by doing nothing).
#
# The default is for the atom to do nothing, when we come 
# across a linkline atom that should have a brain
# we derive a class with one.
#
# Please don't use stuff that starts with an underscore
# it's private.
package Link_Line_Atom; @Link_Line_Atom::ISA = qw(Linked_List_Elem); {
    
sub new {
    my $self = bless {}, shift;
    $self->{_data} = shift;
    return $self;
}

# Give textual representation.
sub get_text {
    my $self = shift;
    return $self->{_data};
}

# Add self to the link-line object.
# Do nothing by default.
# Return: a list of things Link_parser should know about.
# NOTE: if you don't want to return anything (most times) do an explicit return
# otherwise the last thing evaluated in the function is returned.
sub add_self {
	return;
}

sub flush {
    return;
}

} # Link_Line_Atom

package Link_Line_Atom_Unparsed; @Link_Line_Atom_Unparsed::ISA = qw(Link_Line_Atom); {
    
sub new {
    my $self = Link_Line_Atom::new(@_);
    return $self;
}

} # Link_Line_Atom

1;
