#!/bin/env mwperl

# Represents the link line, this class contains information
# about the state of the linked object while traversing the 
# physical link line.

my $La_Dir = "$ENV{MWHOME}/lib/la";
do "$La_Dir/Symbol_Table.pl";
do "$La_Dir/Link_Line_Atom.pl";
do "$La_Dir/Object_Module.pl";
do "$La_Dir/SObject_Module.pl";
do "$La_Dir/Archive_Module.pl";
do "$La_Dir/U_Flag.pl";
do "$La_Dir/B_Flag.pl";
do "$La_Dir/Priority.pl";
do "$La_Dir/Directed_Graph.pl";
do "$La_Dir/Lib_Table.pl";
do "$La_Dir/Symbol_Table.pl";

# Please don't use stuff that starts with an underscore
# it's private.
use strict;
package Link_Parser; {
sub new {
    my $self = bless {}, shift;

    # All symbols
    $self->{_sym_tbl} = Symbol_Table->new();
    # Only symbols in the final module (not from .so files)
    $self->{_mod_tbl} = Symbol_Table->new();
    
    # hash of new unresolved symbols added
    $self->{_new_unres} = {};

    # array of symbols the def file said should be exported.
    $self->{_deffile_exports} = shift;
    @{$self->{_cpp_exports}} = ();

    # array of symbols that should be initially unresolved.
    $self->{_initial_unresolved} = [];
    
    # make a list of symbols that are alway defined.
    $self->_set_up_reserved();

    # Parsed command line (list of objects).
    $self->{_parsed_line} = Link_Line_Atom->new(""); # empty list.

    # list of all modules (shared or static) that were created.
    # prevent us from making more than one object for each module.
    # key = stripped name.
    $self->{_created_modules} = {};

    # list of libraries added by the link analyzer to the link line
    # key: stripped name. value: list of lib objects.
    $self->{_libs_that_were_added} = {};

    $Link_Parser::B_Status = "dynamic";

    $self->{force_archive} = {};

    return $self;
}

# Discard current state so we can start anew.
sub _flush {
    my $self = shift;
    my $hard = shift;

    # All symbols
    $self->{_sym_tbl}->flush();
    $self->{_sym_tbl}->add_defined($self->{_reserved});
    # Only symbols in the final module (not from .so files)
    $self->{_mod_tbl}->flush();
    $self->{_mod_tbl}->add_defined($self->{_reserved});
    
    # Symbols that were recently added.
    %{$self->{_new_unres}} = ();

    # all these symbols are needed.
    $self->add_unresolved($self->{_initial_unresolved});
    
    # keep track of priorities (for AIX)
    @Link_Parser::All_Priorities = ();

    # Keep track of all used mainwin so files for dip.pl
    # key: stripped lib name. value: lib object.
    $self->{_mainwin_sos} = {};

    # Hash of symbols defined by unused libraries
    # Key = symbol name. Value = first defining libraries.
    $self->{_defined_unused} = {};

    # Hash of symbols defined by unused archives good for symbols required
    # by def files, we need them in the final module, a so isn't good enough
    # Key = symbol name. Value = first defining libraries.
    $self->{_defined_unused_archives} = {};

    # Keep track of used libraries and required ones
    $self->{_import_lib} = Lib_Table->new();
    # Don't treat same library twice.
    while( my($name, $libs) = each %{$self->{_libs_that_were_added}} ) {
	$self->{_import_lib}->saw_lib($name, $libs);
    }
    if ($hard) {
	print STDERR "Performing hard flush\n" if $linker_analyzer::Full_Trace;

	my $elem = $self->{_parsed_line};
	do {
	    $elem->flush();
	} while ($elem = $elem->next_elem());
    }
}

sub _add_absent {
    my $self = shift;

    # quick escape, maybe there aren't any:
    my @req_libs = $self->_absent_libraries();
    return 0 if( ! @req_libs );

    # we want to add all required libs to the end of the list
    # find the end of the list
    my $elem;
    my $next = $self->{_parsed_line};
    while($next) {
	$elem = $next;
	$next = $elem->next_elem();
    }

    my %located;
    do {

	my $libname = $req_libs[0];
	
	if (! $located{$libname} ) {
	    $located{$libname} = 1;

	    if( $linker_analyzer::Partial_Trace ) {
		print STDERR "Restarting build with added ",
		"library ($libname)\n";
	    }
	    
	    # key: stripped name, value: lib objects
	    my %lib_objs;
	    my @libs = $self->_locate_libs($libname);
	    $lib_objs{$libname} = [@libs];
	    foreach(@libs) {
		$elem = $elem->append($_);
		$self->_add_elem($elem);
	    }

	    # remember we already treated these libraries and don't 
	    # do so again.
	    foreach my $lib (keys %lib_objs) {
		$ {$self->{_libs_that_were_added}}{$lib} = $lib_objs{$lib};
	    }
	}

	$self->{_import_lib}->delete_absent($libname);
	@req_libs = $self->_absent_libraries();

    } while ( @req_libs);

    return 1;
}

sub add_deffile_symbols {

    my $self = shift;
    my $changed = 0;

    # Check if the deffile requires any library that we skipped 
    # to be added.
    while(my ($unre, $lib) = $self->_find_skipped_archive() ) {
	$changed = 1;
	# if the symbol is defined by a shared library 
	# _find_skipped_lib will fail and this means we need to 
	# move the archive back to before the defining shared 
	# library.
	if( (not $LA_Flags::Do_Not_Reorder) and 
	   (not $self->_find_skipped_lib( $unre ) ) ) {
	    # since we don't know which library defined the symbol
	    # we move the archive to before the first shared 
	    # library on the link line.
	    my $elem = $self->{_parsed_line};
	    $elem = $elem->next_elem() until $elem->isa("SObject_Module");
	    if( $linker_analyzer::Partial_Trace ) {
		print STDERR "Moving archive ",$lib->stripped_name(),
		" towards to back ",
		"of the linkline because of symbol $unre\n";
	    }
	    $elem->prepend( $lib->remove() );
	}
	if( $linker_analyzer::Full_Trace ) {
	    print STDERR "Deffile forced library ",$lib->stripped_name(),
	    " to be added ($unre)\n";
	}

	my $uflag = U_Flag->new($unre, $self);
	#make sure the symbol is unresolved now
	$self->{_sym_tbl}->force_unresolved($unre);
	$self->{_mod_tbl}->force_unresolved($unre);
	
	$lib->set_added($uflag);

	# Don't check _add_elem's return code, we've already
	# changed something.
	$self->_add_elem($lib);
    }
    return $changed;
}

sub _compat_501 {
    my $self = shift;
    return 0 if $LA_Flags::Do_Not_Reorder;
    # maybe order is correct?
    my $changed = 0;
    my $state = 0;
    my $elem = $self->{_parsed_line};
    do {
	if ( ($elem->isa('SObject_Module') or ($elem->isa('Archive_Module'))
	      and
	      ($elem->is_empty_stripped() or $elem->is_system()))) {
	    if ($state > 4) {
		$changed = 1;
	    } else {
		$state = 4;
	    }
	} elsif ( $elem->isa('SObject_Module') ) {
	    if ($state > 3) {
		$changed = 1;
	    } else {
		$state = 3;
	    }
	} elsif ( $elem->isa('Archive_Module') ) {
	    if ($state > 2) {
		$changed = 1;
	    } else {
		$state = 2;
	    }
	} elsif ( $elem->isa('L_Flag') ) {
	    if ($state > 1) {
		$changed = 1;
	    } else {
		$state = 1;
	    }
	} elsif( $elem->isa('Object_Module')) {
	    if ($state > 0) {
		$changed = 1;
	    }
	}
    } while ($changed eq 0 and ($elem = $elem->next_elem()) );
    return 0 if $changed eq 0;

    # Move empty .so file forwards.
    $self->_push_x_forwards(sub { 1 }, sub { shift()->isa('Object_Module') },
			    "compat 501 - objects");
    $self->_push_x_forwards(sub { 1 }, sub { shift()->isa('L_Flag') }, 
			    "compat 501 - -L flag");
    $self->_push_x_forwards(sub { 1 }, sub { shift()->isa('Archive_Module') }, 
			    "compat 501 - Archive Module");
    $self->_push_x_forwards(sub { 1 }, sub { shift()->isa('SObject_Module') },
			    "compat 501 - SObject Module");
    $self->_push_x_forwards(sub { 1 }, sub{ ($_[0]->isa('SObject_Module') or ($_[0]->isa('Archive_Module'))
	      and
					     ($_[0]->is_empty_stripped() or $_[0]->is_system()))}, "compat 501 - stripped libraried");

    return 1;
}

sub _check_force_archive {
    my $self = shift;
    my $changed = 0;
    my $symbol;
    
    foreach $symbol (keys %{$self->{force_archive}}) {

	# If the symbol is already taken from an Object or Archive,
	# we are done.
	next if $self->{_mod_tbl}->is_defined($symbol); # already exists

	# If we don't know of such symbol, there is nothing we can do.
	next if not $self->{_sym_tbl}->is_defined($symbol);

	# OK, so we took the symbol from a SO instead of a archive.
	# Lets see if there is a relevant archive
	my $lib = $self->{_defined_unused_archives}->{$symbol};

	my $elem;
      
	for ($elem = $self->{_parsed_line};($elem = $elem->next_elem() );) {
	    last if ($elem->isa('SObject_Module') and
		     $elem->resolves_symbol($symbol));
	} 

	if ($linker_analyzer::Full_Trace) {
	    my $archname = $lib->stripped_name();
	    my $soname = $elem->stripped_name();
	    print STDERR "Moving archive $archname before shared library " .
		"$soname so that $symbol will be resolved from archive\n";
	}
	$elem->prepend($lib->remove());

    }
    return $changed;
}

sub build_line {
    my $self = shift;
    my $loops =0;
    my $changed = 1;
    my $hard_flush = 0;
    while($changed) { # Iteration loop (pass over the whole link line)
	$changed = 0;
	++$loops;
	$self->_flush($hard_flush); # clean slate
	$hard_flush = 0;

	print STDERR "Starting new iteration\n" 
	    if $linker_analyzer::Full_Trace;

	# Process current known link line, performing some changes
	# on the way
	for( my $elem = $self->{_parsed_line}; $elem; 
	    $elem = $elem->next_elem() ) {
	    $changed = 1 if( $self->_add_elem($elem) );
            print STDERR "In build_line. Adding $elem->{_data}\n" 
		    if $linker_analyzer::Full_Trace; 
	}
	my $tmp_changed = $changed;

	print STDERR "After adding elements, changed = $changed\n"
	    if $linker_analyzer::Full_Trace;

	# Check if the deffile requires any library that we skipped 
	# to be added.
	$changed = 1 if $self->add_deffile_symbols();

	print STDERR "After deffile symbols, changed = $changed\n"
	    if $linker_analyzer::Full_Trace;

	# Add needed libraries that did not appear on the link line
	$changed = 1 if $self->_add_absent();

	print STDERR "After absent libraries, changed = $changed\n"
	    if $linker_analyzer::Full_Trace;

	# restart link analysis if we needed to change something
	# due to topologial sort.
	$changed = 1 if $self->_sort_topologicaly();

	print STDERR "After topological sort, changed = $changed\n"
	    if $linker_analyzer::Full_Trace;

	if ($LA_Flags::Compliant) {

	    $changed = 1 if $self->_check_force_archive();
	    print STDERR "After _check_force_archive, changed = $changed\n"
		if $linker_analyzer::Full_Trace;

	} else {

	    $changed = 1 if $self->_compat_501();

	    print STDERR "After compat501, changed = $changed\n"
		if $linker_analyzer::Full_Trace;
	}


	$hard_flush = 1 if ($changed ne $tmp_changed);
    } # End iteration loop

    if ($linker_analyzer::Full_Trace) {
	print STDERR "Final result of module's table\n";
	$self->{_mod_tbl}->print();
    }

    # Insert undefined symbols for each mainwin shared object
    # that is used.
    my @unres = ();
    foreach my $so ( $self->get_mainwin_sos() ) {
	my $safe = ::calc_safe_name($so);
	my $symbol = "Mw___".$safe."_DependancyNode";
	push(@unres, $symbol);
    }
    $self->add_unresolved(\@unres);
    
    # if we need priority but no Priority object was created.
    if( $linker_analyzer::Check_Priority and not $Priority::priority_used ) {
	my $elem = $self->{_parsed_line};
	# push priority before first object file (or lib)
        $elem = $elem->next_elem() until $elem->isa("Object_Module"); 
	my $priority = Priority->new();
	$elem->prepend($priority);
    }
}

# Return: Was something changed that might require another iteration
# over the link line?
sub _add_elem {
    my ($self, $elem) = @_;

    my $changed = 0; # return value

    my @rc = $elem->add_self($self);
	
    if( @rc ) { # add_self returned something, treat it acordingly.
	if( $ENV{MWOS} eq 'linux' and $elem->isa('Object_Module') and
	   not $LA_Flags::Do_Not_Reorder ) {
	    if($self->_versioned_symbols_fix($elem, @rc) ) {
		$changed = 1;
	    }
	}
    }


    my @unres = keys %{$self->{_new_unres}};
    %{$self->{_new_unres}} = (); # treated current unresolved.

    foreach my $unre (@unres) {

	my ($sym, $lib) = $self->_find_skipped_lib($unre);

	next if not defined $sym;

	# Don't resolve backwards when a dependent shared 
	# library depends on an archive.
	next if $lib->isa("Archive_Module") and $elem->isa("SObject_Module");

	$changed = 1;
	if( $linker_analyzer::Full_Trace ) {
	    print STDERR "Using previously skipped library ", $lib->stripped_name(),
	    " because of symbol $sym.\n";
	}

	# Archives need a -u flag and SOs don't
	my $uflag;
	if( $lib->isa("Archive_Module") ) {
	    $uflag = U_Flag->new($sym, $self); 
		# Make sure symbol is unresolved now
		$self->{_sym_tbl}->force_unresolved($sym);
		$self->{_mod_tbl}->force_unresolved($sym);
	}
	$lib->set_added($uflag);

	$self->_add_elem($lib);
    }

    return $changed;
}


# get textual version of the link line to be passed to the 
# native linker.
sub get_link_line {
    my $self = shift;
    %Link_Parser::added_lib_dirs = (); # Set of -L flags added for libraries in funny places
    my @ret = ();
    for( my $elem = $self->{_parsed_line}; $elem; $elem = $elem->next_elem() ) {
	my $txt = $elem->get_text();
	push(@ret, $txt) if $txt; # get rid of empty strings
    }
    return join(" ", @ret);
}

sub _locate_libs {
    my ($self, $name) = (shift, shift);
    
    my @ret = ();
    my $located = linker_analyzer::locate_lib($name, $Link_Parser::B_Status);
    foreach my $lib ( @$located ) {
	my $stripped = $lib->{stripped_name};
	# don't create same module twice.
	next unless $self->_should_create_module( $lib );

	if( $lib->{type} eq $linker_analyzer::Archive_t ) {
	    my $obj = Archive_Module_L->new($stripped,
					    $lib->{source},
					    $Link_Parser::B_Status, $lib);
	    push(@ret, $obj);
	}
	elsif( $lib->{type} eq $linker_analyzer::Sharedlib_t ) {
	    my $obj = SObject_Module_L->new($stripped, 
					    $lib->{source},
					    $Link_Parser::B_Status, $lib);
	    push(@ret, $obj);
	}
	elsif ( $linker_analyzer::Corrupted_Libs{$stripped} ) {
	    # Couldn't locate library, passing to ld as is.
	    my $obj = Corrupt_Module->new("-l".$stripped);
            print STDERR "W: _locate_libs - Corrupt_Module created. $obj $obj->{_data}\n" if $linker_analyzer::Full_Trace;
	    push(@ret, $obj);
	}
    }
    return @ret;
}

# This is method parses the textual command line
# and creates an array of corresponding Link_Line_Atoms.
# It uses the global variable $Link_Parser::B_Status which can
# be either `dynamic' or `static'.
sub parse_command_line {
    my ($self, @command_line) = @_;

    # default value
    $Link_Parser::B_Status = "dynamic";

    # build a list with objects for each element in command_line
    my $parsed = $self->{_parsed_line};
    while ( @command_line ) {
	my $flag = shift @command_line;
	next unless defined $flag;
	my @elems = $self->_parse_element( $flag,\@command_line);
	foreach my $elem (@elems) {
	    $parsed = $parsed->append($elem);
	}
    }
  Object_Module::init_all_o_files();
}

# Return a list of created objects (usually one but can be zero or many).
sub _parse_element {
    my ($self, $flag, $command_line) = @_;

    if ( $flag =~ /\.o$/ ) {
	my $obj = Object_Module->new($flag);
	return ($obj);
    }
    if ( $flag =~ /\.a$/ ) { # explicitly added archive
	my $lib_obj = linker_analyzer::process_lib($flag);
	if( $lib_obj->{type} ne $linker_analyzer::Archive_t ) {
	    # file wasn't located, pass as is to ld.
	    return Link_Line_Atom_Unparsed->new($flag);
	}
	if( $self->_should_create_module( $lib_obj ) ) {
	    my $obj = Archive_Module->new($flag, $Link_Parser::B_Status, $lib_obj);
	    return ($obj);
	}
	return; # if no object created return no value
    }
    if ( $flag =~ /\.s[ol]$/ ) { # explicitly added shared object
	my $lib_obj = linker_analyzer::process_lib($flag);
	if( $lib_obj->{type} ne $linker_analyzer::Sharedlib_t ) {
	    # file wasn't located, pass as is to ld
	    return Link_Line_Atom_Unparsed->new($flag);
	}
	if( $self->_should_create_module( $lib_obj ) ) {
	    my $obj = SObject_Module->new($flag, $Link_Parser::B_Status, $lib_obj);
	    return ($obj);
	}
	return; # if no object created return no value
    }
    # lib files (added with -l)
    if( $flag =~ s/^-l// ) {
	$flag = shift @$command_line if $flag =~ /^\s*$/;
	return $self->_locate_libs($flag);
    }
    if( $flag =~ s/^-u\s*// ) {
	if( $flag =~ /^\s*$/ ) {
	    $flag = shift @$command_line;
	}
	my $obj = U_Flag->new($flag, $self);
	return ($obj);
    }
    if( $flag =~ /^-B/ ) {
	my $regexp = "(static|dynamic)\$";
	if($flag =~ /^-B\s*$/ ) {
	    if( $$command_line[0] =~ /^$regexp/ ) {
		$flag = shift @$command_line;
	    }
	} 
	if( $flag =~ /^(-B)?\s*$regexp/ ) {
	    $flag =~ s/^-B\s*//;
	    my $obj = B_Flag->new($flag);
        $Link_Parser::B_Status = $obj->{_b_status};
	    return ($obj);
	}
    }
    if( $flag =~ s/^-L// ) {
	if( $flag =~ /^\s*$/ ) {
	    $flag = shift @$command_line;
	}
	my $obj = L_Flag->new($flag);
	return ($obj);
    }
    if( $flag =~ s/^-o// ) {
	if( $flag =~ /^\s*$/ ) {
	    $flag = shift @$command_line;
	}
	my $obj = Minus_O->new($flag);
	return ($obj);
    }
    if( $linker_analyzer::Check_Priority and $flag =~ /^-p(.*)/ ) {
	$flag = $1;
	if( $flag =~ /^\s*$/ ) {
	    $flag = shift @$command_line;
	}
	my $obj = Priority->new($flag,"-p");
	return ($obj);
    }
    if( $linker_analyzer::Check_Priority and $flag =~ /^-qmkshrobj(.*)/ ) {
	$flag = $1;
	if( $flag =~ /^=(.*)$/ ) {
	    $flag = $1;
	} else {
	    $flag = $Priority::Default_Priority;
	}
	my $obj = Priority->new($flag,"-qmkshrobj=");
	return ($obj);
    }
    if( $flag eq "-mwstartas_is" || $flag =~ /^-mwas_is\"/) {
	my @asis;
	
	# added in a comma delimited string.
	if( $flag =~ /^-mwas_is\"(.*)\"/ ) {
	    @asis = split(",", $1);
	} 
	else { # added until -mwendas_is
	    while( ($flag = shift @$command_line) ne "-mwendas_is" ) {
		push(@asis, $flag);
	    }
	}

	# convert x.lib to -lx
	s/^(.*?)\.lib$/-l$1/ foreach(@asis);
	
	my $obj = Link_Line_Atom_Unparsed->new(join(" ", @asis));
	return ($obj);
    }
    if ( $flag =~ /^-(no)?defaultlib/ ) {
	return;
    }

    my $obj = Link_Line_Atom_Unparsed->new($flag);
    return ($obj);
}

# gets a full path to a library and returns the stripped
sub strip_library_name {
    my $path = shift;
    my ($pre, $name, $ext);
    ::split_path($path, \$pre, \$name, \$ext);
    # get rid of initial lib
    $name =~ s/^lib//;
    return $name;
}

# an unneded library was encountered
sub unneeded_lib {
    my ($self, $lib) = @_;
    my $unused = $self->{_defined_unused};
    my $unused_arch = $self->{_defined_unused_archives};

    my $is_arch = $lib->isa("Archive_Module");
    local $^W = 0; # We're referencing undefs here.
    # ||= means assign if the current value is false (undef)
    foreach my $def ($lib->get_unused_defined()) {
	$unused->{$def} ||= $lib;
	if( $is_arch ) {
	    $unused_arch->{$def} ||= $lib;
	}
    }
}

# Add unresolved symbols 
sub add_unresolved {
    my ($self, $syms) = @_;
    $self->{_mod_tbl}->add_unresolved($syms);
    $self->{_sym_tbl}->add_unresolved($syms);
}

# Merge a symbol table from static source (archive or .o file).
# Params: 
#	1) Other symbol table.
#	2) Optional, should we check if the merge is needed, defaults to false.
#	3) Optional, reference to an array for the symbols that were already 
#		defined.
# Returns: was the merge actually made. 
sub merge_static {
    my ($self, $table, $conditional, $old_defs, $reason) = @_;

    # Do the same as merge_dynamic
    my ($dyn_old, $static_old);
    if( $old_defs ) {
	$dyn_old = [];
	$static_old = [];
    }
    my $result = $self->merge_dynamic($table, $conditional, $dyn_old, $reason);
    if ($result eq 1) {
	# and also add the the final module's symbol table if we do
	# the merge.
	# Note, we don't use $conditional, if merge_dynamic did a merge
	# we want the static merge too.
	# We don't want to know about new undefiens so the second params
	# is undef
	$self->{_mod_tbl}->merge_symbol_table($table, undef, 0, $static_old);
	if( $old_defs ) {
	    @$old_defs = (@$dyn_old, @$static_old);
	}
    } elsif ($result eq 2) {
	$self->{_mod_tbl}->merge_symbol_table($table, undef, 1, $static_old);
    }

    return $result;
}

# Merge a symbol table from a dynamic library
# Params: See merge static.
# Side effect: updates _new_unres
sub merge_dynamic {
    my ($self, $table, $conditional, $old_defs, $reason) = @_;
    
    # We want to get the new unres.
    my @new_unres = ();
    my $ret = $self->{_sym_tbl}->merge_symbol_table($table, \@new_unres,
						    $conditional, $old_defs,
						    $reason);
    if( $ret and @new_unres ) {
	@{$self->{_new_unres}}{@new_unres} = (1)x @new_unres;
    }

    return $ret;
}
    
    
# See if any of the libraries we decided not to add became needed.
# Return: a two elemented list (symbol, defining-library) or an empty
# list if not found.
sub _find_skipped_lib {
    my $self = shift;

    foreach my $unre ( @_ ) {
	# Only interested in symbols that aren't defined
	next if $self->{_sym_tbl}->is_defined($unre);
	# Check unused libs
	my $lib = $self->{_defined_unused}->{$unre};
	if( $lib and not $lib->provides_symbol($unre) ) {
	    return ($unre, $lib);
	}
    }
    return ();
}

# See if any of the libraries we decided not to add became needed.
# Return: a two elemented list (symbol, defining-library) or an empty
# list if not found.
sub _find_skipped_archive {
    my $self = shift;
    return unless %$self->{_deffile_exports};

    foreach my $unre (keys %{$self->{_deffile_exports}}) {
	# Only interested in symbols that aren't defined in the final module
	next if $self->{_mod_tbl}->is_defined($unre);
	# Check unused archives
	my $arch = $self->{_defined_unused_archives}->{$unre};
	if( $arch and not $arch->provides_symbol($unre) ) {
	    return ($unre, $arch);
	}
    }
    return ();
}


# reserved symbols are always resolved
sub _set_up_reserved {
    my $self = shift;

    if( $ENV{MWOS} eq "sunos5" ) {
	$self->{_reserved} = [qw(_GLOBAL_OFFSET_TABLE_)];
    }
    elsif ( $ENV{MWOS} =~ /^ux1[10]$/ ) {
	$self->{_reserved} = [qw(_main __curStaticObject)];
    }
    else { 
	$self->{_reserved} = [];
    }
}


# Return a list of all the defined symbols in the final
# module (doesn't include so files the module depends on)
sub get_globals_in_module {
    my $self = shift;
    return $self->{_mod_tbl}->get_defined();
}

# Return a list of all the unresolved symbols in the final
# module (doesn't include so files the module depends on)
sub get_undefs_in_module {
    my $self = shift;
    return $self->{_mod_tbl}->get_unresolved();
}

# Return whether the symbol is defined in the final module (see above).
sub defined_in_module {
    my ($self, $symbol) = @_;
    return $self->{_mod_tbl}->is_defined($symbol);
}

# Return a list of all the unresolved symbols (including stuff from
# dynamic libraries).
sub get_unresolved {
    my $self = shift;
    return $self->{_sym_tbl}->get_unresolved();
}

# Add symbols to the initial_unresolved field.
sub force_unresolved {
    my $self = shift;
    push(@{$self->{_initial_unresolved}}, @_);
}

# Log the existence of a library on the link line.
sub saw_library {
    my $self = shift;
    my ($name, $libs) = (shift, shift);
    unless ( ref($libs) eq 'ARRAY' ) {	$libs = [$libs]; }
    $self->{_import_lib}->saw_lib($name, $libs);
}

# Some libraries are required by pragma comment lib or dependencies.
sub require_libraries {
    my $self = shift;
    # remove trailing .lib s
    s/\.lib$// foreach @_;
    if ( $linker_analyzer::Use_Libmap ) {
	my @tmp;
	foreach (@_) {
	    my $mapped = $::libmap{$_};
	    if( not defined $mapped ) {
		push(@tmp, $_); # No mapping keep original
	    }
	    elsif ( $mapped and $mapped ne '@' ) {
		# @ means ignore and "" means mapped to nothing.
		push(@tmp, split(" ", $mapped) );
	    }
	}
	@_ = @tmp;
    }

    $self->{_import_lib}->require_libs(@_);
}

# input: A symbol table. 
# Find all the symbols that indicate that a #pragma comment(lib,...) was
# used and require those libraries.
sub require_libs_from_pragma_comment_lib {
    my ($self, $table) = @_;
    # The format is __mw_link_<number>_<string>_<number>_lib_<lib name>
    my @safe = grep(s/__mw_link_(?:\d+)_(?:.*?)_(?:\d+)_lib_(.*)/$1/, $table->get_defined());
    my @safe2 = grep(s/__mw_link_(?:\d+)_(?:.*?)_(?:\d+)_lib_(.*)/$1/, $table->get_common());
    push (@safe, @safe2);
    # translate them back to their real "unsafe" names.
    my @libs = map( linker_analyzer::uncalc_safe_name($_), @safe );
    $self->require_libraries(@libs);

    my @safeflags = grep(s/__mw_link_(?:\d+)_(?:.*?)_(?:\d+)_linker_(.*)/$1/, $table->get_defined());
    my @safeflags2 = grep(s/__mw_link_(?:\d+)_(?:.*?)_(?:\d+)_linker_(.*)/$1/, $table->get_common());
    push (@safeflags, @safeflags2);
    my @flags = map( linker_analyzer::uncalc_safe_name($_), @safeflags );
    my $flag;
    foreach $flag (@flags) {
	if ($flag =~ /^-force-archive (.*)/) {
	    $self->{force_archive}->{$1} = 1;
	}
    }
}

# get list of all required libs that aren't on the linkline
sub _absent_libraries {
    my $self = shift;
    return $self->{_import_lib}->get_absent();
}

# A mainwin shared object is being used.
sub saw_mainwin_so {
    my ($self, $so) = @_;
    $self->{_mainwin_sos}->{$so->{stripped_name}} = $so;
}

# get list of all the mainwin shared object files.
sub get_mainwin_sos {
    my $self = shift;
    return keys %{$self->{_mainwin_sos}};
}

# Create a DAG of the shared objects
sub _build_dag {
    my $self = shift;
    my $graph = Directed_Graph->new();
    for(my $elem = $self->{_parsed_line}; $elem; $elem = $elem->next_elem()) {
	if( $elem->isa("SObject_Module") and $elem->is_added() ) {
	    my @depends = $elem->dependencies();
	    next if not @depends;
	    foreach my $depname (@depends) {
		my $deps = $self->{_import_lib}->get_value($depname);
		if( $deps ) {
		    foreach my $dep (@$deps) {
			$graph->add_edge($elem, $dep) unless _does_not_depend($elem, $dep);
		    }
		} 
		else {
		    print STDERR "Library ", $elem->stripped_name(),
		    " depends on unknown library ($depname).\n";
		}
	    }
	}
    }
    $graph->dagify();
    return $graph;
}

# returns whether any changes were made.
sub _sort_topologicaly {
    return if $LA_Flags::Do_Not_Reorder;

    my $self = shift;
    my $dag = $self->_build_dag();

    my $changed_something = 0;
    my $elem = $self->{_parsed_line};
    while( $elem ) {
	my $dep = ($dag->get_dependents($elem))[0];
	# No action needed, tag and move on.
	if( not $dep ) {
	    $dag->tag_node($elem);
	    $elem = $elem->next_elem();
	    next;
	}

	# Moving stuff around...
	if( $linker_analyzer::Partial_Trace ) {
	    print STDERR "Topological sort moving ", $elem->get_text(), 
	                 " in front of ", $dep->get_text(),".\n";
	}

	$changed_something = 1;

	# get next element in iteration
	my $tmp = $elem->next_elem();
	$elem->remove(); # remove from list
	$dep->append($elem); # and add in front of dependant
	$elem = $tmp;
    }
    # Move empty .so file forwards.
    $self->_push_x_forwards(sub { shift()->isa('SObject_Module')},
			    \&SObject_Module::is_empty_stripped, "empty" );
    # Move system libs in front of all .so and .a files
    $self->_push_x_forwards(sub { $_[0]->isa('SObject_Module') or $_[0]->isa('Archive_Module') },
			    \&SObject_Module::is_system, "system");
    # Move all corrupt .so files forwards.
    $self->_push_x_forwards(sub { $_[0]->isa('SObject_Module') or $_[0]->isa('Archive_Module') },
			    sub { shift()->isa('Corrupt_Module') }, "unknown");
    # Move unparsed tokens
    $self->_push_x_forwards(sub { $_[0]->isa('SObject_Module') or $_[0]->isa('Archive_Module') or $_[0]->isa('Link_Line_Atom_Unparsed')},
			    sub { $_[0]->isa('Link_Line_Atom_Unparsed') and $_[0]->get_text() =~ /^:/ }, "Unparsed token" );

    # We don't care if system empty or currupt libraries changed location 
    # because they don't change our outlook anyway.
    return $changed_something;
}


# move all libs that satisfy function x forwards.
sub _push_x_forwards {
    my ($self, $relevant, $tester, $description) = @_;
    my $elem = $self->{_parsed_line};
    my $last_non_x = $elem; # in case there are no such SOs
    do {
	if( $elem->$relevant() and not $elem->$tester() ) {
	    $last_non_x = $elem;
	}
    } while ($elem = $elem->next_elem() );

    my $changed_something = 0;
    $elem = $self->{_parsed_line};
    my $curr = $last_non_x;
    while($elem != $last_non_x) {
	if( $elem->$relevant() and $elem->$tester() ) {
	    print STDERR "Moving $description library ",$elem->get_text()," forward.\n" 
		if $linker_analyzer::Partial_Trace;
	    $changed_something = 1;
	    my $next = $elem->next_elem();
	    $curr->append( $elem->remove() );
	    $curr = $elem;
	    $elem = $next;
	}
	else { $elem = $elem->next_elem(); }
    }
    
    return $changed_something;
}

# This method checks if a $module has been created
# and assumes that if it hadn't been created up until now it will be
# created after the function is called. This means for every uniqe name
# it returns true the first time it's called and false every following time.
sub _should_create_module {
    my ($self, $module) = @_;
	my $fullname = "undefined";
	if ($module->{source} ne "") {
	$fullname = "$module->{source}";
	}
	else {
	$fullname = "$module->{stripped_name}";
	}

    # already exists, return false
    return if( $self->{_created_modules}->{$fullname} ); 
    # didn't exist, return true.
    $self->{_created_modules}->{$fullname} = 1;
    return 1;
}

# We may get some false dependencies because of the libmap.
# In this method we determine if we have prior knowldge that this link should 
# not be made.
# params: $from: dependant node, $to: to node
sub _does_not_depend {
    my ($from, $to) = @_;
    # We only have problems if we're using libmap
    return unless $linker_analyzer::Use_Libmap;

    my $fromname = $from->stripped_name();
    my $toname = $to->stripped_name();
    
    # treat self dependencies
    if( $fromname eq $toname ) { 
	return 1; 
    }

    if( $toname eq 'gdiuser32' and $fromname eq 'advapi32' ) { 
	return 1; 
    }

    return;
}

# on linux if we're building a dll and one of the symbols defined in an
# .o file or an archive have already been defined in a previous dll
# the linker will fail. To solve this we take all the dlls that come before
# this element and push them forward if they define a symbol from the list 
# we got.
sub _versioned_symbols_fix {
	my ($self, $target, @syms) = @_;
	my $changed = 0;
# loop until the element in question
	for( my $elem = $self->{_parsed_line}; $elem != $target;
			$elem = $elem->next_elem()) {
		# only deal with DLLs
		if( $elem->isa('SObject_Module') ) {
			# check all the syms we got
			foreach my $sym (@syms) {
				# if one is found move the dll in front of our target
				if( $elem->resolves_symbol($sym)  ) {
					print STDERR "Moving DLL ",$elem->stripped_name(),
					" in front of $target->{_data} for fear of clashing ",
					"versioned symbol $sym.\n" if $linker_analyzer::Full_Trace;

					my $tmp = $elem;
					# Save the place for the for loop
					$elem = $elem->previous_elem();
					$target->append($tmp->remove());
					$changed = 1;
					last;
				}
			}
		}
	}
	return $changed;
}

# In case of 'strong++' map support, there are exported symbols with
# C++-like syntax. Find them and put them in a special table.
sub extract_cpp_exports {
    my $self = shift;
    foreach my $expsym (keys %{$self->{_deffile_exports}}) {
	if ($expsym =~ /^(.+)::(.+)$/) {
	    push(@{$self->{_cpp_exports}},"$1\\S*$2");
	    push(@{$self->{_cpp_exports}},"$2\\S*$1");
	    delete ${$self->{_deffile_exports}}{$expsym};
	}
    }
    return;
}

sub add_exports {
    my ($self, $exps) = @_;
    @{$self->{_deffile_exports}}{@$exps} = (1)x @$exps;
}

} # Link_Parser
1;
