#!/bin/env mwperl

# Represents an object file, the base class represents a .o
# file and derived classes represent .so and .a.
do "$ENV{MWHOME}/lib/la/Symbol_Table.pl";
do "$ENV{MWHOME}/lib/la/Link_Line_Atom.pl";

use strict;

# Platform specific nm flags and stuff
package NM_Flags;
sub new { bless {}, shift; }

# Init platform specific stuff related to nm 
sub _init_nm_stuff {
    my $self = shift;
    
    # flags to pass to nm
    my $default_nm_flags = " -A -g -p ";
    my %nm_flags = (
		    sunos5 => $default_nm_flags,
		    aix4   => " -A -p -P -g -C ",
		    linux  => " -A -P -g ",
		    ux10   => $default_nm_flags,
		    ux11   => " -A -P -g ",
		    );
    $self->{_nm_flags} = $nm_flags{$ENV{MWOS}};

    # regexp for finding a defined symbol
    my $default_nm_regexp_d = "^[TDBA]";
    my %nm_regexp_d = (
		       sunos5 => $default_nm_regexp_d,
		       aix4   => $default_nm_regexp_d,
		       linux  => "^[DTWBRV]",
		       ux10   => "^[TDABC]",
		       ux11   => "^[TDABCRS]"
		       );
    $self->{_nm_regexp_d} = $nm_regexp_d{$ENV{MWOS}};
    
    $self->{_nm_regexp_u} = "^U";

    # permutation to order the nm output line in canonical form
    my @ord_ux  = qw(1 2);
    my %nm_order = (
		    sunos5 => [qw(2 3)],
		    aix4   => [qw(2 1 0)],
		    linux  => [qw(2 1)],
		    ux10   => \@ord_ux,
		    ux11   => [qw(2 1)],
		    );
    $self->{_nm_order} = $nm_order{$ENV{MWOS}};

    $self->{_nm_cmd} = ($ENV{MWOS} eq 'linux')? '/usr/bin/nm' : '/usr/ccs/bin/nm';
}

package Object_Module; @Object_Module::ISA = qw(Link_Line_Atom); {
# Please don't use stuff that begins with an underscore
# it's private.

    # Accumulate all the .o files so that we can call nm
    # once for all of them instead of for every single file.
    %Object_Module::_all_o_files = ();

sub new {
    my $self = Link_Line_Atom::new(@_);
    print STDERR "new Object_Module, $self $self->{_data}\n" if $linker_analyzer::Full_Trace;
    $Object_Module::_all_o_files{$self->{_data}} = $self;
    return $self;
}

sub name {
    my $self = shift;
    return $self->{_data};
}

# Returns the name of the file that resolves this symbol [true] 
# (in an archive it is differnt than the name of the module) 
# or "" [false] if the symbol is unresolved.
sub resolves_symbol {
    my ($self, $symbol) = @_;
    unless ( defined $self->{_sym_tbl} ) {
        print STDERR "\tE: resolves_symbol: symbol table missing for $self\n" if $linker_analyzer::Full_Trace;
	return "";
    }
    if( defined $symbol and $self->{_sym_tbl}->is_defined($symbol) ) {
	return $self->name();
    }
    return "";
}

# Returns: is this symbol supplied by this module?
# The module needs to be used and the symbol defined.
sub provides_symbol {
    my ($self, $symbol) = @_;
    if( defined $symbol and $self->is_added() ) { 
	return $self->resolves_symbol($symbol);
    }
    return;
}

# Add this module to a link line parser
# The default behaviour (in an .o file) is to add the 
# symbol table to both the global and final symbol tables, 
# in deriving classes (.so and .a) more work needs to be done.
# Returns: a list of all the defined symbols that were already defined
#	we need this for linux because of versioned symbols.
sub add_self {
    my ($self, $parser) = @_;
    return if (!defined $self->{_sym_tbl});
    $parser->require_libs_from_pragma_comment_lib($self->{_sym_tbl});
    # Treat common symbols on Sun
    $self->_add_common_symbols($parser);

    # On linux we need to keep track of the defined symbols
    # that were already defined
    my $old_syms;
    if( $ENV{MWOS} eq 'linux' ) {
	$old_syms = [];
    }
    # Back to business (Note second params is false, always use .o files).
    $parser->merge_static( $self->{_sym_tbl}, 0, $old_syms  );
    return $old_syms? @$old_syms : undef;
}

# in Sun give common symbols special treatment.
sub _add_common_symbols {
    my ($self, $parser) = (shift, shift);

    # add my common symbols unless they are already defined
    for my $sym (keys %{$self->{_common}}) {
	$linker_analyzer::Common_Syms{$sym} = 1 unless $parser->defined_in_module($sym);
    }
    # delete my defined non-common symbols from the common symbols
    foreach my $def ( $self->{_sym_tbl}->get_defined() ) {
	# don't create values by mistake.
	if( exists $self->{_common}->{$def} and $self->{_common}->{$def} ) { next; }

	# else: defined and not common
	delete $linker_analyzer::Common_Syms{$def};
    }
}

# Return whether this object module is used in the 
# final link, .o files return true. Deriving classes 
# for .so and .a files need to do some work here.
sub is_added {
    return 1;
}

# If the module isn't added we return an empty string
# otherwise we return its representation
sub get_text {
    my $self = shift;
    if( $self->is_added() ) {
	return $self->Link_Line_Atom::get_text();
    }
    return "";
}

# Input: symbol data in an array in the form "[UD] $symbol"
# Output: a new Symbol_Table object.
# NOTE: there are three very similar methods if one changes 
# chances are all should change
# _build_sym_table
# _sunos5_build_o_file_sym_table
# SObject_Module::_build_sym_table
sub _build_sym_table {
    my $self = shift;

    my @defs = ();
    my @unres = ();
    my @common = ();
    foreach my $line (@_ ) {
	my ($ty, $info) = split(' ', $line);
	
	if( $ty eq 'D' ) {
	    push(@defs, $info);
	}
	elsif( $ty eq 'U' ) {
	    push(@unres, $info);
	}
	elsif( $ty eq 'C' ) {
	    $self->{_common}->{$info} = 1;
	    push(@common, $info);
	}
    }

    my $sym_tbl = Symbol_Table::new Symbol_Table;
  
    $sym_tbl->add_defined(\@defs);
    $sym_tbl->add_unresolved(\@unres);
    $sym_tbl->add_common(\@common);

    foreach my $pattern (@{$linker_analyzer::link_parser->{_cpp_exports}}) {
	my @exps = grep /$pattern/,@defs;
	$linker_analyzer::link_parser->add_exports(\@exps) if (@exps);
    }

    return $sym_tbl;
}



# Note these are static functions.
# Get a filename and the data to fill the symbol table with.
sub _init_o_file {
    my ($file, @data) = @_;
    unless ($file) {
	    print STDERR "\tERROR: _init_o_file called with empty file argument\n" if $linker_analyzer::Full_Trace;
	    return;
    }
    my $obj = $Object_Module::_all_o_files{$file};
    die "Can't find object for $file" 
	unless eval { $obj->isa('Object_Module') };
    
    $obj->{_sym_tbl} = $obj->_build_sym_table(@data);

    print STDERR "\tINFO: _init_o_file : object is $obj , symbol table is $obj->{_sym_tbl}\n" if $linker_analyzer::Full_Trace;
}

sub init_all_o_files {
    # Create dummy NM_Flags object so we have the nm stuff.
    my $nm_data = NM_Flags->new();
    $nm_data->_init_nm_stuff();

    # create string of .o files, each surronded by quotes,
    # in order to support spaces in .o file names.
    my $list = 	'"'.join("\" \"",keys %Object_Module::_all_o_files).'"';

    if ( $ENV{MWOS} eq 'sunos5' ) {
	# treat common symbols
	_sunos5_init_all_o_files($nm_data, $list);
	return;
    }

    my $cmd = $nm_data->{_nm_cmd}.$nm_data->{_nm_flags}." $list |";
    
    print STDERR "\t$cmd\n" if $linker_analyzer::Full_Trace;
	
    open(NM, $cmd) or die "Can't run nm on object files: $!";

    my $pattern = '(^[^:*]*):';
    if ($ENV{MWCONFIG_NAME} eq 'ia64_ux11') {
	$pattern = '([^:\s]*):';
    }

    my $prev_file = '';
    my @data = ();
    while(<NM>) {
	chomp;
	#skip empty lines or lines that only contain "file:"
	next if (/^[^:]*:\s*$/ or /^\s*$/);

	my $line = $_;
	# There is a file name and it's new
	if( ($line =~ /$pattern/o) and (my $new_file = $1) ne $prev_file) {
	    # On UX11 if nm gets an empty .o file its output doesn't fit our patterns.
	    if( $ENV{MWOS} eq 'ux11' and 
	       $ENV{MWCONFIG_NAME} ne 'ia64_ux11' and
	       $line =~ /^nm:[^:]*:\s*no symbols/) {
		next;
	    }

	    _init_o_file($prev_file, @data) if $prev_file;
            
            # start treating new file.
	    $prev_file = $new_file;
	    @data = ();
	}	    
	my @words = split(' ', $line);
	my ($ty, $info) = @words[@{$nm_data->{_nm_order}}];

	if( $ty =~ /$nm_data->{_nm_regexp_d}/o ) {
	    next if $ENV{MWOS} eq "aix4" and $info =~/^\./o;
	    push( @data, "D ".$info);
	}
	elsif ( $ty =~ /$nm_data->{_nm_regexp_u}/o) {
	    if( $ENV{MWOS} eq "aix4" ) {
		$info =~ s/^\.//; # Remove intial dot if exists
	    }
	    push(@data, "U ".$info);
	}
    }
    close(NM);
    _init_o_file($prev_file, @data) if $prev_file; # treat last .o file
}

# This is a special case of _init_all_o_files for sunos5
# it should be used only of .o files so that we treat 
# common symbols correctly.
sub _sunos5_init_all_o_files {
    my ($nm_data, $list) = @_;
    
    my $cmd = $nm_data->{_nm_cmd}." -A $list |";
    print STDERR "\t$cmd\n" if $linker_analyzer::Full_Trace;
    open(NM, $cmd) or die "Can't run nm on object files: $!";
    
    my $prev_file = '';
    my @data = ();
    while(<NM>) {
	chomp;

	# skip empty lines
	next if /:\s*$/o or  /^\s*$/;
	next unless /GLOB|WEAK/;

	my @words = split (/\|/, $_);
	my $info = $words[7];

	next unless $info;

	# There is a file name and it's new
	if( /^([^:]*):/ and (my $new_file = $1) ne $prev_file) {
	    _init_o_file($prev_file, @data) if $prev_file;
            # start treating new file
	    $prev_file = $new_file;
	    @data = ();
	}	    

	if ( $words[6] =~ /UNDEF/ ) { 
	    push(@data, "U ".$info);
	}
	else {
	    # remember common symbols
	    if( $words[6] =~ /COMMON/ ) {
		# sun's _build_sym_table knows how to deal with this.
		push(@data, "C ".$info);
	    } 
	    else {
		push(@data, "D ".$info);
	    }
	}
    }
    close NM;
    _init_o_file($prev_file, @data) if $prev_file; # treat last .o file
}

} # End package Object_Module

# Some modules care if they're linked dynamicly or statically.
# Keep track of the situation for them.
package B_Status_Aware; @B_Status_Aware::ISA = qw(Link_Line_Atom); {

sub set_b_status {
    my $self = shift;
    $self->{_b_status} = scalar(@_)? shift : "dynamic";
    # do we need to change the parser's state?
    $self->{_b_explicit} = 0;
}

sub get_b_status {
    my $self = shift;
    my $ret = $self->{_b_status};
    if( not(defined $ret) or not $ret ) {
	$ret = "dynamic";
    }
    return $ret;
}


sub add_self {
    my ($self, $parser) = @_;
    
    my $status = $self->get_b_status();
    $self->{_b_explicit} = ($status ne $Link_Parser::B_Status);
    if( $self->{_b_explicit} ) {
	# We need to change the global state
	$Link_Parser::B_Status = $status;
    }
	return;
}

sub get_text {
    my $self = shift;
    if( $self->{_b_explicit} ) {
	return "-B".$self->get_b_status()." ";
    }
    return "";
}


} # B_Status_Aware
 
1; # Make use succeed
