#!/bin/env mwperl

# Represents a symbol table.
# This class allows building a symbol table (by correctly
# managing the addition of defined and unresolved symbols)
# and getting the current state of the symbol table.
# It is also possible to merge two symbol tables or more.
use strict;
package Symbol_Table; {
# Please don't use stuff that begins with an underscore
# it's private.

sub new {
    my $self = bless {}, shift;
    my @ignores = ();
    $self->{_defined} = {};
    $self->{_unresolved} = {};
    $self->{_common} = {};
    $self->{_common_resolves} = 
	($ENV{MWOS} eq 'sunos5' || $ENV{MWOS} eq 'linux');

    if( $ENV{MWOS} eq 'sunos5' ) {
    	@ignores = ('__1c2K6Fpv_v_', '__1c2N6FI_pv_', '__1c2k6Fpv_v_', '__1c2n6FI_pv_');
    }
    elsif($ENV{MWOS} eq 'linux') {
	@ignores = ('_Znaj', '_ZdaPv', '_Znwj', '_ZdlPv'); 
    }
    elsif($ENV{MWOS} eq 'ux11') {
	@ignores = ('__nwa__FUl', '__nw__FUl', '__dl__FPv', '__dla__FPv');
    }
    elsif($ENV{MWOS} eq 'aix4') {
	@ignores = ('__vd__FPv', '__vn__FUl', '__dl__FPv', '__nw__FUl');
    }
	 
    foreach my $sym (@ignores) {
    	$self->{_ignored}->{$sym} = 1;	
    }
    return $self;
}


# check if a specific symbol is defined
sub is_defined {
    my ($self, $symbol) = @_;
    return exists $self->{_defined}->{$symbol};
}

# check if any of the symbols is defined
# the list of symbols should be passed by reference
# Return: in array context the list of all the defined symbols
#         in scalar context true if any of the symbols are defined
#         false otherwise.
sub any_defined {
    my ($self, $symbols) = @_;
    my $defined = $self->{_defined};

    # Note: Duplicated code, I don't want to have an if in every loop
    # so we check once in advance.
    if( wantarray ) {
	my @ret = ();
	foreach my $sym (@$symbols) {
	    push(@ret, $sym) if ( $sym and not exists $self->{_ignored}->{$sym} and $defined->{$sym});
	}
	return @ret;
    }
    else {
	foreach my $sym (@$symbols) {
	    return 1 if ($sym and not exists $self->{_ignored}->{$sym} and $defined->{$sym});
	}
    }
    return;
}
  

# check is specific symbol is unresolved
sub is_unresolved {
    my ($self, $symbol) = @_;
    return exists $self->{_unresolved}->{$symbol};
}

# check is specific symbol is ignored
sub is_ignored {
    my ($self, $symbol) = @_;
    return exists $self->{_ignored}->{$symbol};
}

sub get_ignored {
    my $self = shift;
    return keys %{$self->{_ignored}};
}

sub get_defined {
    my $self = shift;
    return keys %{$self->{_defined}};
}

sub get_unresolved {
    my $self = shift;
    return keys %{$self->{_unresolved}};
}

sub get_common {
    my $self = shift;
    return keys %{$self->{_common}};
}


# Add defined symbols
# parameters:
#    1) reference to a list of symbols
#    2) (optional) conditional add? (defaults to false)
#    3) (optional) reference to an array to be filled with
#       symbols that were already defined.
# Returns: Was the addition made?
sub add_defined {
    my ($self, $defs, $conditional, $old_syms, $reason) = @_;

    my $defined = $self->{_defined};
    my $unres = $self->{_unresolved};

    my $deleted = 0;
    foreach my $symbol (@$defs) {
	if( $old_syms and $defined->{$symbol} ) {
	    push @$old_syms, $symbol; # symbol was already defined
	    next; # This iteration is redundant
	}
	# if this symbol was unresolved it no longer is.
	if (delete $unres->{$symbol} && !$deleted) {
	    $deleted = 1;
	    $$reason = $symbol if $reason;
	}
    }

    # Do the actual addition if it's not a conditional merge
    # or the condition was met.
    if( (not $conditional) or $deleted ) {
	@{$defined}{@$defs} = (1) x @$defs;
	return 1;
    }
    return '';
}

# Add unresolved symbols
# The list of symbols must be passed by reference.
# Return the newly unresolved symbols
sub add_unresolved {
    my ($self, $symbols) = @_;
    
    my $defined = $self->{_defined};
    my $unres = $self->{_unresolved};

    my @new_unres = ();
    foreach my $symbol (@$symbols) {
	if( not exists $defined->{$symbol} ) {
	    push(@new_unres, $symbol) unless $unres->{$symbol};
	    $unres->{$symbol} = 1;
	}
    }
    return @new_unres;
}

# Add common symbols
# The list of symbols must be passed by reference.
# Return the newly unresolved symbols
sub add_common {
    my ($self, $symbols, $promote_only, $reason) = @_;
    
    my $defined = $self->{_defined};
    my $unres = $self->{_unresolved};
    my $common = $self->{_common};

    my @new_common = ();
    if ($promote_only) {
	foreach my $symbol (@$symbols) {
	    if ($unres->{$symbol}) {
		delete $unres->{$symbol};
		push(@new_common, $symbol);
		$common->{$symbol} = 1;
		$$reason = $symbol if ($reason);
	    }
	}
    } else {
	foreach my $symbol (@$symbols) {
	    if( not exists $defined->{$symbol} ) {
		if ($unres->{$symbol}) {
		    delete $unres->{$symbol};
		    push(@new_common, $symbol);
		} else {
		    push(@new_common, $symbol) unless $unres->{$symbol};
		}
		$common->{$symbol} = 1;
	    }
	}
    }
    return @new_common;
}

# Force a symbol to be unresolved even if it was defined before
# Return, was it defined before?
sub force_unresolved {
	my ($self, $symbol) = @_;

	$self->{_unresolved}->{$symbol} = 1;
	delete $self->{_common}->{$symbol};
	return delete $self->{_defined}->{$symbol};
}

# Merge the other symbol table into this one.
# Params: 
#     1) Other symbol table
#     2) Optional reference to list for newly unresolved symbols
#     3) Optional flag to do a conditional merge
#     4) Optional reference to list for defined symbols that were
#        already defined.
# Return: was the merge made?
sub merge_symbol_table {
    my ($self, $other, $new_unres, $cond, $old_syms, $reason) = @_;

    my $result = '';

    # In any case, we want to promote the common symbols
    $result = 2 if $self->add_common( [$other->get_common()], $cond, $reason );

    if( $self->add_defined( [$other->get_defined()], $cond, 
			   $old_syms, $reason) ||
       ($self->{_common_resolves} && $result eq 2)) {
	@$new_unres = $self->add_unresolved( [$other->get_unresolved()] );
	$self->add_common( [$other->get_common()] ) if ($cond);
	return 1;
    }

    return $result;
}

# flush the symbol table
sub flush {
    my $self = shift;
    %{$self->{_unresolved}} = ();
    %{$self->{_defined}} = ();
}

sub print {
    my $self = shift;
    print STDERR "Defined: ";
    my $symbol;
    foreach $symbol ($self->get_defined()) {
	print STDERR "$symbol ";
    }
    print STDERR "\nUnresolved: ";
    foreach $symbol ($self->get_unresolved()) {
	print STDERR "$symbol ";
    }
    print STDERR "\nCommon: ";
    foreach $symbol ($self->get_common()) {
	print STDERR "$symbol ";
    }
    print STDERR "\nIgnored: ";
    foreach $symbol ($self->get_ignored()) {
	print STDERR "$symbol ";
    }
    print STDERR "\n";

}

} # End package Symbol_Table

# Make use succeed
1;
